/* File: A-mac-h.c */

/*
 * Purpose: Source for the pre-compiled header "A-mac-h"
 * This file is used by the Think C 6.0 Macintosh version
 *
 * Open this file from inside Think C 6.0, and select "precompile..."
 * from the "source" menu and save the result as "A-mac-h".  Then you
 * can do "Bring up to date" from the "Project" menu.  Note that the
 * file "A-mac-h" is included by the project "prefix" section.
 */

/*
 * Activate the Macintosh-specific code
 */
#define MACINTOSH

/*
 * OPTION: Allow the Borg
 */
/* #define ALLOW_BORG */

/*
 * OPTION: Allow wizard commands
 */
/* #define ALLOW_WIZARD */

/*
 * OPTION: Generate spoilers
 */
/* #define ALLOW_SPOILERS */

/*
 * Precompile the header files
 */
#include "angband.h"

