/* File: borg-ben.h */

/* Purpose: Header file for "borg-ben.c" -BEN- */

#ifndef INCLUDED_BORG_BEN_H
#define INCLUDED_BORG_BEN_H

#include "angband.h"

#ifdef ALLOW_BORG

/*
 * This file provides support for "borg-ben.c".
 */


/*
 * Interact with the Ben Borg
 */
extern void borg_ben(void);

/*
 * Initialize this file
 */
extern void borg_ben_init(void);


#endif

#endif

